﻿using System;
using System.ComponentModel;
using Loki.Core.Test.Events;
using Loki.Engine;
using Loki.Engine.Registration;
using Loki.UI;
using Loki.Utils;
using NUnit.Framework;

namespace Loki.Core.Test
{
    [TestFixture]
    public class LokiServicesTest : AssertionHelper
    {
        [SetUp]
        public void SetUp()
        {
            ToolKit.Reset();
        }

        [Test]
        public void LoggerPresenceTest()
        {
            Expect(
                ToolKit.Logger,
                Is.Not.Null
                );
        }

        [Test]
        public void CommandPresenceTest()
        {
            Expect(
                ToolKit.Command,
                Is.Not.Null
                );
        }

        [Test]
        public void NoTrackingTest()
        {
            ObjectModel L_Model = new TestObjectModel();
            L_Model.Initialize();

            WeakReference L_ObjectReference = null;
            new Action(() =>
            {
                object L_Object = L_Model.Create<INotifyPropertyChanged>();
                Expect(L_Object, Is.Not.Null);
                Expect(L_Object.GetType(), Is.EqualTo(typeof(DummyObserved)));
                L_ObjectReference = new WeakReference(L_Object);
            })();

            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();

            Assert.IsNull(L_ObjectReference.Target);
        }

        [Test]
        public void ObjectModelTest()
        {
            ObjectModel L_Model = new TestObjectModel();
            L_Model.Initialize();
            object L_Object = L_Model.Create<INotifyPropertyChanged>();
            Expect(L_Object, Is.Not.Null);

            Expect(L_Object.GetType(), Is.EqualTo(typeof(DummyObserved)));
        }

        [Test]
        public void ViewModelCleanUpTest()
        {
            ObjectModel L_Model = new TestObjectModel();
            ToolKit.Engine.DefineUIModel(L_Model);
            L_Model.Initialize();

            WeakReference L_Reference = null;
            new Action(() =>
            {
                IViewModel VW = ToolKit.Engine.GetViewModel<TestViewModel>();
                L_Reference = new WeakReference(VW);
            })();

            Expect(L_Reference.Target, Is.Not.Null);
            ((IViewModel)L_Reference.Target).DoCleanup();
            GC.Collect();
            GC.WaitForPendingFinalizers();
            GC.Collect();

            Expect(L_Reference.Target, Is.Null);
        }

        [Test]
        public void RedefinitionTest()
        {
            ToolKit.Context.Register(Element.For<IEngine>().ImplementedBy<DummyEngine>());
            Expect(ToolKit.Errors, Is.Not.Null);
            Expect(ToolKit.Event, Is.Not.Null);
            Expect(ToolKit.Logger, Is.Not.Null);
            Expect(ToolKit.Engine as DummyEngine, Is.Not.Null);
            Expect(ToolKit.Task, Is.Not.Null);

            Expect(
                ToolKit.Command,
                Is.Not.Null
                );
        }

        [Test]
        public void SingletonTest()
        {
            var L_Service1 = ToolKit.Logger;
            var L_Service2 = ToolKit.Logger;
            Expect(L_Service1 == L_Service2, Is.True);
        }

        [Test]
        public void OverrideTest()
        {
            using (IObjectContext L_Context = ToolKit.Engine.CreateContext("Test"))
            {
                L_Context.Register(Element.For<ISupportInitialize>().ImplementedBy<DummyComponent1>());
                L_Context.Register(Element.For<ISupportInitialize>().Named("Test").ImplementedBy<DummyComponent2>());

                L_Context.Register(Element.For<DummyRoot>().Overrides(Loki.Engine.Registration.Property<DummyRoot>.ForKey(x => x.Component).Value("Test")));

                DummyRoot L_Root = L_Context.Resolve<DummyRoot>();
                Expect(L_Root.Component, Is.InstanceOf<DummyComponent2>());
            }
        }

        [Test]
        public void OverrideDoubleNameTest()
        {
            using (IObjectContext L_Context = ToolKit.Engine.CreateContext("Test"))
            {
                L_Context.Register(Element.For<IEngine>().ImplementedBy<DummyEngine>().Named("Test"));
                L_Context.Register(Element.For<ISupportInitialize>().Named("Test").ImplementedBy<DummyComponent2>());

                IEngine L_Engine = L_Context.Resolve<IEngine>("Test");
                ISupportInitialize L_Init = L_Context.Resolve<ISupportInitialize>("Test");

                Expect(L_Engine, Is.Not.Null);
                Expect(L_Init, Is.Not.Null);
            }
        }
    }
}