﻿using System;
using System.Globalization;
using System.Linq;
using Loki.Commands;
using Loki.Engine;
using Loki.Engine.Registration;
using Loki.Utils;

namespace Loki.UI
{
    public class BootStrapperAdapter<TMainViewModel> : LokiObject
        where TMainViewModel : class, IViewModel
    {
        protected IApplication Application { get; private set; }

        public BootStrapperAdapter(IApplication P_Application)
        {
            try
            {
                Application = P_Application;
                Commands = new CommandManager();

                Commands.Handle(ApplicationCommands.START, ApplicationStart_Execute);
                Commands.Handle(ApplicationCommands.EXIT, ApplicationExit_Execute);
                Commands.Handle(ApplicationCommands.MESSAGE, ApplicationMessage_Execute);
                Commands.Handle(ApplicationCommands.WARNING, ApplicationWarning_Execute);
                Commands.Handle(ApplicationCommands.ERROR, ApplicationError_Execute);

                Application.LastChanceError += Application_UnhandledException;
            }
            catch (Exception L_Ex)
            {
                Log.Error(L_Ex.Message, L_Ex);
                Commands.Dispose();
                throw;
            }
        }

        protected CommandManager Commands { get; private set; }

        #region Initializing

        public event EventHandler Initializing;

        protected virtual void OnInitializing(EventArgs e)
        {
            EventHandler L_Handler = Initializing;

            if (L_Handler != null)
            {
                L_Handler(this, e);
            }
        }

        #endregion Initializing

        #region Initialized

        public event EventHandler Initialized;

        protected virtual void OnInitialized(EventArgs e)
        {
            EventHandler L_Handler = Initialized;

            if (L_Handler != null)
            {
                L_Handler(this, e);
            }
        }

        #endregion Initialized

        public void Application_UnhandledException(object P_Sender, ExceptionEventArgs e)
        {
            Application.Error(e.Exception, true);
            ExitApplication();
        }

        public void Initialize()
        {
            OnInitializing(EventArgs.Empty);

            try
            {
                // Define templating engine
                ToolKit.Context.Register(Element.For<ITemplatingEngine>().Instance(Application.Templates));

                // Initialize Data engine
                ToolKit.Engine.DefineDataModel(Application.DataModel);
                Application.DataModel.Initialize();

                // Initialize application engine
                ToolKit.Engine.DefineApplicationModel(Application.ApplicationModel);
                Application.ApplicationModel.Initialize();

                // Initialize ui engine
                ToolKit.Engine.DefineUIModel(Application.UIModel);
                Application.UIModel.Initialize();

                OnInitialized(EventArgs.Empty);
                PreStart();
            }
            catch (Exception L_Ex)
            {
                Application.Error(L_Ex, true);
                ExitApplication();
            }
        }

        public virtual void PreStart()
        {
            ApplicationCommands.START.Execute(null);
        }

        public void Start()
        {
            // Creates main objects
            TMainViewModel L_MainModel = ToolKit.Engine.GetViewModel<TMainViewModel>();
            Application.MainView = ToolKit.Templates.GetView<TMainViewModel>(L_MainModel);

            // activates main form
            Application.MainView.Show();
        }

        public void ExitApplication()
        {
            Application.ApplicationExit();
        }

        #region Commands

        private void ApplicationStart_Execute(object sender, CommandEventArgs e)
        {
            Start();
        }

        private void ApplicationExit_Execute(object sender, CommandEventArgs e)
        {
            ExitApplication();
        }

        private void ApplicationMessage_Execute(object sender, CommandEventArgs e)
        {
            Application.Message(string.Format(CultureInfo.CurrentUICulture, Resources.Formats.Message_Format, e.Parameter));
        }

        private void ApplicationWarning_Execute(object sender, CommandEventArgs e)
        {
            Application.Warning(string.Format(CultureInfo.CurrentUICulture, Resources.Formats.Message_Format, e.Parameter));
        }

        private void ApplicationError_Execute(object sender, CommandEventArgs e)
        {
            Application.Error(e.Parameter as Exception, false);
        }

        #endregion Commands

        #region ViewModels

        /// <summary>
        /// Gets the view models.
        /// </summary>
        public LokiCollection<IViewModel> ViewModels
        {
            get;
            private set;
        }

        protected TViewModel ActivateOrAdd<TViewModel>(Func<TViewModel, bool> P_Selector, Action<TViewModel> P_Initializer) where TViewModel : class,IViewModel
        {
            // search
            TViewModel L_Model = default(TViewModel);
            if (P_Selector != null)
            {
                L_Model = ViewModels.OfType<TViewModel>().FirstOrDefault(x => P_Selector(x));
            }
            else
            {
                L_Model = ViewModels.OfType<TViewModel>().FirstOrDefault();
            }

            // init
            if (L_Model == null)
            {
                L_Model = AddNew<TViewModel>(P_Initializer);
            }
            else
            {
                L_Model.Activate.Execute(null);
            }

            return L_Model;
        }

        protected TViewModel AddNew<TViewModel>(Action<TViewModel> P_Initializer) where TViewModel : class,IViewModel
        {
            TViewModel L_NewModel = ToolKit.Engine.GetViewModel<TViewModel>();
            if (P_Initializer != null)
            {
                P_Initializer(L_NewModel);
            }

            AddSubModel(L_NewModel);
            return L_NewModel;
        }

        private void AddSubModel(IViewModel P_Model)
        {
            Commands.Handle(P_Model.Close, (s, e) => SubModel_CanClose(P_Model, e), (s, e) => SubModel_Close(P_Model, e));
            ViewModels.Add(P_Model);

            IView L_View = ToolKit.Templates.GetView(P_Model);
            if (L_View != null)
            {
                L_View.Show();
            }
        }

        private void SubModel_CanClose(IViewModel P_Model, CanExecuteCommandEventArgs e)
        {
            e.CanExecute = true;
        }

        private void SubModel_Close(IViewModel P_Model, CommandEventArgs P_Args)
        {
            Commands.UnHandle(P_Model.Close);
            P_Model.DoCleanup();
            ViewModels.Remove(P_Model);
        }

        #endregion ViewModels
    }
}