﻿using System;
using System.Diagnostics;
using Loki.Commands;
using Loki.Engine;

namespace Loki.Utils
{
    /// <summary>
    /// Loki services helper class.
    /// </summary>
    public static class ToolKit
    {
        #region Extensibility

        internal static LokiExtensionManager Extensions { get; private set; }

        static ToolKit()
        {
            Extensions = new LokiExtensionManager();
            Extensions.Initialize();
            _Context = new Lazy<IObjectContext>(CreateContext);
        }

        #endregion Extensibility

        #region Object Context

        private static Lazy<IObjectContext> _Context;

        public static IObjectContext Context
        {
            get
            {
                return _Context.Value;
            }
        }

        public static void Reset()
        {
            if (_Context.IsValueCreated)
            {
                Context.Dispose();
                Console.WriteLine("Context dispoded");
            }

            _Context = new Lazy<IObjectContext>(CreateContext);
        }

        private const string LOKI_CONTEXT = "Loki";

        private static IObjectContext CreateContext()
        {
            try
            {
                var L_Context = Extensions.Engine.CreateContext(LOKI_CONTEXT);
                new LokiServicesInstaller().Install(L_Context);
                Extensions.Engine.Initialize();
                return L_Context;
            }
            catch (Exception L_Ex)
            {
                Trace.Fail(L_Ex.Message);
                throw;
            }
        }

        #endregion Object Context

        /// <summary>
        /// Loads the additional context.
        /// </summary>
        /// <param name="P_Installer">The context installer.</param>
        public static void LoadContext(IContextInstaller P_Installer)
        {
            P_Installer.Install(Context);
        }

        public static ITemplatingEngine Templates
        {
            get
            {
                return Context.Resolve<ITemplatingEngine>();
            }
        }

        /// <summary>
        /// Gets the engine service.
        /// </summary>
        public static IEngine Engine
        {
            get
            {
                return Context.Resolve<IEngine>();
            }
        }

        /// <summary>
        /// Gets the logger service.
        /// </summary>
        /// <value>The configured logger service.</value>
        public static ILogService Logger
        {
            get
            {
                return Context.Resolve<ILogService>();
            }
        }

        /// <summary>
        /// Gets the settings service.
        /// </summary>
        /// <value>The configured settings service.</value>
        public static ISettingsService Settings
        {
            get
            {
                return Context.Resolve<ISettingsService>();
            }
        }

        /// <summary>
        /// Gets the command service.
        /// </summary>
        /// <value>The configured command service.</value>
        public static ICommandService Command
        {
            get
            {
                return Context.Resolve<ICommandService>();
            }
        }

        /// <summary>
        /// Gets the event service.
        /// </summary>
        public static IEventService Event
        {
            get
            {
                return Context.Resolve<IEventService>();
            }
        }

        /// <summary>
        /// Gets the task service.
        /// </summary>
        public static ITaskService Task
        {
            get
            {
                return Context.Resolve<ITaskService>();
            }
        }

        /// <summary>
        /// Gets the error service.
        /// </summary>
        public static IErrorService Errors
        {
            get
            {
                return Context.Resolve<IErrorService>();
            }
        }
    }
}