﻿using System;
using System.Threading;
using System.Windows.Forms;
using Loki.Commands;
using Loki.Engine;
using Loki.UI.Win.Resources;
using Loki.Utils;

namespace Loki.UI.Win
{
    public class WinFormBootStrapper<TMainViewModelType> : ApplicationContext, IApplication
        where TMainViewModelType : class, IViewModel
    {
        protected BootStrapperAdapter<TMainViewModelType> BootStrapper { get; private set; }

        private BootForm _BootForm;

        #region Log

        private ILog _Log = null;

        /// <summary>
        /// Gets the logger.
        /// </summary>
        /// <value>The logger.</value>
        protected ILog Log
        {
            get
            {
                if (_Log == null)
                {
                    Interlocked.CompareExchange(ref _Log, ToolKit.Logger.GetLogger(this.GetType().Name), null);
                }

                return _Log;
            }
        }

        #endregion Log

        #region Run

        public static void Run()
        {
            Run(new WinFormBootStrapper<TMainViewModelType>());
        }

        public static void Run(WinFormBootStrapper<TMainViewModelType> P_Boot)
        {
            Application.SetCompatibleTextRenderingDefault(false);

            // If using EnableVisualStyles, always call DoEvents immediately afterwards. Failing to
            // do so can result in an unexpected SEHException.
            Application.EnableVisualStyles();
            Application.DoEvents();

            P_Boot.BootStrapper.Initialize();

            Application.Run(P_Boot);
        }

        #endregion Run

        protected WinFormBootStrapper()
        {
            BootStrapper = new BootStrapperAdapter<TMainViewModelType>(this);

            Application.ThreadException += new ThreadExceptionEventHandler(Application_ThreadException);
            BootStrapper.Initializing += BootStrapper_Initializing;
            BootStrapper.Initialized += BootStrapper_Initialized;
        }

        private void BootStrapper_Initialized(object P_Sender, EventArgs e)
        {
            _BootForm.Close();
            _BootForm.Dispose();
        }

        private void BootStrapper_Initializing(object P_Sender, EventArgs e)
        {
            _BootForm = new BootForm();
            _BootForm.Label = Messages.UIMessages.APP_START;
            _BootForm.Show();
            Application.DoEvents();
        }

        private void Application_ThreadException(object P_Sender, ThreadExceptionEventArgs e)
        {
            OnLastChanceError(new ExceptionEventArgs(e.Exception));
        }

        public IView MainView
        {
            get { return MainForm as IView; }

            set { MainForm = value as Form; }
        }

        private void MainModel_CanClose(object sender, CanExecuteCommandEventArgs e)
        {
            e.CanExecute = true;
            Log.Debug("Main ViewModel closing ask");
        }

        private void MainModel_Close(object sender, CommandEventArgs e)
        {
            Log.Debug("Main Viewmodel closing");
        }

        public void ApplicationExit()
        {
            if (_BootForm != null)
            {
                _BootForm.Dispose();
            }

            Application.Exit();
        }

        private ITemplatingEngine _Templates = new UnityWinformEngine();

        public virtual ITemplatingEngine Templates
        {
            get { return _Templates; }
        }

        private ObjectModel _ApplicationModel = new ObjectModel();

        public virtual ObjectModel ApplicationModel
        {
            get { return _ApplicationModel; }
        }

        private ObjectModel _DataModel = new ObjectModel();

        public virtual ObjectModel DataModel
        {
            get { return _DataModel; }
        }

        private ObjectModel _UIModel = new ObjectModel();

        public virtual ObjectModel UIModel
        {
            get { return _UIModel; }
        }

        #region LastChanceError

        public event EventHandler<ExceptionEventArgs> LastChanceError;

        protected virtual void OnLastChanceError(ExceptionEventArgs e)
        {
            EventHandler<ExceptionEventArgs> L_Handler = LastChanceError;

            if (L_Handler != null)
            {
                L_Handler(this, e);
            }
        }

        #endregion LastChanceError

        public void Message(string P_Message)
        {
            MessageBoxes.Message(P_Message);
        }

        public void Warning(string P_Warning)
        {
            MessageBoxes.Warning(P_Warning);
        }

        public void Error(Exception P_Exception)
        {
            LokiException L_Ex = P_Exception as LokiException;
            if (L_Ex != null)
            {
                ErrorMessageBox.Show(L_Ex);
            }
            else
            {
                ErrorMessageBox.Show(ToolKit.Errors.BuildError<LokiException>(ErrorMessages.TUI_000, Log, L_Ex));
            }
        }

        public void Error(Exception P_Exception, bool P_Imperative)
        {
            throw new NotImplementedException();
        }
    }
}