﻿using System;
using System.Collections.ObjectModel;
using System.Threading;
using System.Windows;
using Loki.Engine;
using Loki.UI.Wpf.Resources;
using Loki.Utils;

namespace Loki.UI.Wpf
{
    public class BootStrapper : Application, IApplication
    {
        #region Log

        private ILog _Log = null;

        /// <summary>
        /// Gets the logger.
        /// </summary>
        /// <value>The logger.</value>
        protected ILog Log
        {
            get
            {
                if (_Log == null)
                {
                    Interlocked.CompareExchange(ref _Log, ToolKit.Logger.GetLogger(this.GetType().Name), null);
                }

                return _Log;
            }
        }

        #endregion Log

        public BootStrapper()
            : base()
        {
            ViewModels = new ObservableCollection<IViewModel>();
            _Templates = new WindsorWPFEngine();
        }

        /// <summary>
        /// Gets the view models.
        /// </summary>
        public ObservableCollection<IViewModel> ViewModels
        {
            get;
            private set;
        }

        #region Properties

        public string ThemeName { get; set; }

        #endregion Properties

        #region Disposable

        /// <summary>
        /// Releases all resources used by an instance of the <see cref="BootStrapperWPF" /> class.
        /// </summary>
        /// <remarks>
        /// This method calls the virtual <see cref="Dispose(bool)" /> method, passing in
        /// <strong>true</strong>, and then suppresses finalization of the instance.
        /// </remarks>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~BootStrapper()
        {
            Dispose(false);
        }

        /// <summary>
        /// Releases the unmanaged resources used by an instance of the
        /// <see cref="BootStrapperWPF" /> class and optionally releases the managed resources.
        /// </summary>
        /// <param name="P_Disposing">Sets <strong>true</strong> to release both managed and
        /// unmanaged resources; <strong>false</strong> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool P_Disposing)
        {
            ToolKit.Context.Dispose();
        }

        #endregion Disposable

        public void ApplicationExit()
        {
            Dispose();

            Log.Info(LogStrings.BootStrapper_ApplicationExit);
            Application.Current.Shutdown();
        }

        private ITemplatingEngine _Templates ;

        public virtual ITemplatingEngine Templates
        {
            get { return _Templates; }
        }

        private ObjectModel _ApplicationModel = new ObjectModel();

        public virtual ObjectModel ApplicationModel
        {
            get { return _ApplicationModel; }
        }

        private ObjectModel _DataModel = new ObjectModel();

        public virtual ObjectModel DataModel
        {
            get { return _DataModel; }
        }

        private ObjectModel _UIModel = new ObjectModel();

        public virtual ObjectModel UIModel
        {
            get { return _UIModel; }
        }

        #region LastChanceError

        public event EventHandler<ExceptionEventArgs> LastChanceError;

        protected virtual void OnLastChanceError(ExceptionEventArgs e)
        {
            EventHandler<ExceptionEventArgs> L_Handler = LastChanceError;

            if (L_Handler != null)
            {
                L_Handler(this, e);
            }
        }

        #endregion LastChanceError

        public void Message(string P_Message)
        {
            MessageBoxes.Message(P_Message);
        }

        public void Warning(string P_Warning)
        {
            MessageBoxes.Warning(P_Warning);
        }

        public void Error(Exception P_Exception, bool P_Imperative)
        {
            LokiException L_Ex = P_Exception as LokiException;
            if (L_Ex != null)
            {
                ErrorMessageBox.Show(L_Ex, P_Imperative);
            }
            else
            {
                ErrorMessageBox.Show(ToolKit.Errors.BuildError<LokiException>(ErrorMessages.Application_UnhandledException, Log, P_Exception), P_Imperative);
            }
        }

        public IView MainView
        {
            get
            {
                return MainWindow as IView;
            }

            set
            {
                MainWindow = value as Window;
            }
        }
    }
}