﻿using System;
using System.Collections.Concurrent;
using System.Globalization;
using System.Linq;
using Castle.MicroKernel.Registration;
using Castle.Windsor;
using Castle.Windsor.Installer;
using Loki.Engine;
using Loki.Engine.Registration;
using Loki.Utils;

namespace Loki.Windsor
{
    public class WindsorContext : LokiObject, IObjectContext
    {
        public WindsorContext()
        {
            InternalContainer = new WindsorContainer();
        }

        protected WindsorContainer InternalContainer { get; private set; }

        public void LoadContext(params string[] P_ObjectDefinitionURI)
        {
            foreach (var L_Config in P_ObjectDefinitionURI)
            {
                InternalContainer.Install(Configuration.FromXmlFile(L_Config));
            }
        }

        private ConcurrentDictionary<Type, Func<object>> _Builders = new ConcurrentDictionary<Type, Func<object>>();

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~WindsorContext()
        {
            Dispose(false);
        }

        protected virtual void Dispose(bool P_Disposing)
        {
            InternalContainer.Dispose();
        }

        public T Resolve<T>(string P_ObjectName) where T : class
        {
            return InternalContainer.Resolve<T>(GetName(typeof(T), P_ObjectName));
        }

        public T Resolve<T>() where T : class
        {
            var L_Type = typeof(T);
            if (!_Builders.ContainsKey(L_Type))
            {
                return InternalContainer.Resolve<T>();
            }
            else
            {
                return _Builders[L_Type]() as T;
            }
        }

        private const string NAME_REGISTRATION_FORMAT = "{0}|{1}";

        private string GetName(Type P_Type, string P_Name)
        {
            return string.Format(CultureInfo.InvariantCulture, NAME_REGISTRATION_FORMAT, P_Type.FullName, P_Name);
        }

        public void Register<T>(ElementRegistration<T> P_Definition) where T : class
        {
            ComponentRegistration<object> L_Registration = Component.For(P_Definition.Types);

            if (P_Definition.Lifestyle.Type == LifestyleType.NoTracking)
            {
                RegisterNoTracking(P_Definition);
            }

            if (!string.IsNullOrEmpty(P_Definition.Name))
            {
                L_Registration = L_Registration.Named(GetName(typeof(T), P_Definition.Name));
            }

            if (P_Definition.Lifestyle.SingletonInstance != default(T))
            {
                L_Registration.Instance(P_Definition.Lifestyle.SingletonInstance).LifestyleSingleton();
            }
            else
            {
                L_Registration = L_Registration.ImplementedBy(P_Definition.Implementation).IsDefault();

                switch (P_Definition.Lifestyle.Type)
                {
                    case LifestyleType.Singleton:
                        L_Registration = L_Registration.LifestyleSingleton();
                        break;

                    case LifestyleType.Pooled:
                        L_Registration = L_Registration.LifestylePooled(P_Definition.Lifestyle.PoolSizeMin, P_Definition.Lifestyle.PoolSizeMax);
                        break;

                    case LifestyleType.Transient:
                        L_Registration = L_Registration.LifestyleTransient();
                        break;

                    default:
                        break;
                }
            }

            foreach (var L_Property in P_Definition.ConfiguredProperties)
            {
                L_Registration = L_Registration.DependsOn(Castle.MicroKernel.Registration.Property.ForKey(L_Property.Key.Name).Eq(L_Property.Value));
            }

            foreach (var L_Property in P_Definition.ConfiguredOverrides)
            {
                L_Registration = L_Registration.DependsOn(Dependency.OnComponent(L_Property.Key.Name, GetName(L_Property.Key.PropertyType, L_Property.Value.ToString())));
            }

            InternalContainer.Register(L_Registration);
        }

        public void Release(object P_Object)
        {
            InternalContainer.Release(P_Object);
        }

        private void RegisterNoTracking<T>(ElementRegistration<T> P_Definition) where T : class
        {
            Type L_InterfaceType = P_Definition.Types.First();
            if (!_Builders.ContainsKey(L_InterfaceType))
            {
                // build new builder
                var L_Constructor = P_Definition.Implementation.GetConstructor(Type.EmptyTypes);
                Func<object> L_Builder = () => L_Constructor.Invoke(new object[0] { });
                _Builders.TryAdd(L_InterfaceType, L_Builder);
            }
        }
    }
}