﻿using System;
using System.ComponentModel.Composition;
using Loki.Engine;
using Loki.Engine.Registration;
using Loki.Utils;

namespace Loki.Windsor
{
    [Export(typeof(IEngine))]
    [ExportMetadata("Type", "Windsor")]
    public class WindsorEngine : LokiObject, IEngine, IDisposable
    {
        private const string DATA_CONTEXT = "DataContext";
        private const string MODEL_CONTEXT = "ModelContext";
        private const string UI_CONTEXT = "UIContext";

        private IObjectContext _UIContext;
        private IObjectContext _ModelContext;
        private IObjectContext _DataContext;

        public void Initialize()
        {
            _UIContext = CreateContext(UI_CONTEXT);
            _DataContext = CreateContext(DATA_CONTEXT);
            _ModelContext = CreateContext(MODEL_CONTEXT);
        }

        public TDataModel GetDataModel<TDataModel>() where TDataModel : class, IObjectCreator
        {
            return _DataContext.Resolve<ObjectModel>() as TDataModel;
        }

        public TApplicationModel GetApplicationModel<TApplicationModel>() where TApplicationModel : class, IObjectCreator
        {
            return _ModelContext.Resolve<ObjectModel>() as TApplicationModel;
        }

        public TViewModel GetViewModel<TViewModel>() where TViewModel : class, UI.IViewModel
        {
            var L_SelectedVM = _UIContext.Resolve<TViewModel>();
            if (!L_SelectedVM.IsInitialized)
            {
                L_SelectedVM.Initialize();
            }

            L_SelectedVM.Clean += L_SelectedVM_Clean;

            return L_SelectedVM;
        }

        private void L_SelectedVM_Clean(object P_Sender, EventArgs e)
        {
            UI.IViewModel L_Object = P_Sender as UI.IViewModel;
            if (L_Object != null)
            {
                L_Object.Clean -= L_SelectedVM_Clean;
                _UIContext.Release(L_Object);
            }
        }

        public void DefineDataModel(ObjectModel P_Model)
        {
            P_Model.InternalContext = _DataContext;
            _DataContext.Register(Element.For<ObjectModel>().Instance(P_Model));
        }

        public void DefineApplicationModel(ObjectModel P_Model)
        {
            P_Model.InternalContext = _ModelContext;
            _ModelContext.Register(Element.For<ObjectModel>().Instance(P_Model));
        }

        public void DefineUIModel(ObjectModel P_Model)
        {
            P_Model.InternalContext = _UIContext;
            _UIContext.Register(Element.For<ObjectModel>().Instance(P_Model).Lifestyle.Singleton);
        }

        public IObjectContext CreateContext(string P_ContextName)
        {
            return new WindsorContext();
        }

        /// <summary>
        /// Releases all resources used by an instance of the <see cref="WindsorEngine" /> class.
        /// </summary>
        /// <remarks>
        /// This method calls the virtual <see cref="Dispose(bool)" /> method, passing in
        /// <strong>true</strong>, and then suppresses finalization of the instance.
        /// </remarks>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        ~WindsorEngine()
        {
            Dispose(false);
        }

        /// <summary>
        /// Releases the unmanaged resources used by an instance of the <see cref="WindsorEngine" />
        /// class and optionally releases the managed resources.
        /// </summary>
        /// <param name="P_Disposing"><strong>true</strong> to release both managed and unmanaged
        /// resources; <strong>false</strong> to release only unmanaged resources.</param>
        protected virtual void Dispose(bool P_Disposing)
        {
            if (_UIContext != null)
            {
                _UIContext.Dispose();
            }

            if (_DataContext != null)
            {
                _DataContext.Dispose();
            }

            if (_ModelContext != null)
            {
                _ModelContext.Dispose();
            }
        }
    }
}