﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using Loki.Data;
using Loki.Engine;
using Loki.Utils;
using NUnit.Framework;

namespace Luna.Data.Test
{
    public abstract class BaseTest<TService, TEntity, TKey> : AssertionHelper
        where TService : class, IDataProvider
        where TEntity : class, IDataEntity<Guid>
    {
        public TService Service
        {
            get
            {
                return ToolKit.Engine.GetDataModel<TestDataModel>().Get<TService>();
            }
        }

        public BaseTest()
        {
            Randomizer = new Random();
        }

        protected Random Randomizer
        {
            get;
            private set;
        }

        protected const string TEST_PREFIX = "#Test";
        private const string STRING_FORMAT = "{0}_{1}_{2}";
        protected static readonly DateTime BASE_DATE = new DateTime(2100, 1, 1);

        protected string GetString(string P_Name)
        {
            return string.Format(CultureInfo.InvariantCulture, STRING_FORMAT, TEST_PREFIX, Randomizer.Next(1000), P_Name);
        }

        protected string GetString(string P_Name, int P_CharCount)
        {
            var L_BaseString = GetString(P_Name);
            return L_BaseString.Substring(0, Math.Min(P_CharCount, L_BaseString.Length));
        }

        protected int GetInteger()
        {
            return Randomizer.Next();
        }

        protected DateTime GetDate()
        {
            return BASE_DATE.AddDays(Randomizer.Next(1000));
        }

        protected bool GetBoolean()
        {
            return Randomizer.Next(10) % 2 == 0;
        }

        public virtual TEntity Create()
        {
            return ToolKit.Engine.GetDataModel<TestDataModel>().Create<TEntity>();
        }

        public abstract bool AreSame(TEntity P_Entity1, TEntity P_Entity2);

        public abstract Predicate<TEntity> DeletePredicate { get; }

        public abstract Func<TEntity, TKey> KeyGetter { get; }

        public abstract Func<IEnumerable<TEntity>> GetAll { get; }

        public abstract Action<TEntity> Save { get; }

        public abstract Action<TKey> Delete { get; }

        public abstract Func<TKey, TEntity> Get { get; }

        [TearDown]
        public virtual void Clean()
        {
            foreach (var L_Item in GetAll().Where(x => DeletePredicate(x)))
            {
                Delete(KeyGetter(L_Item));
            }
        }

        [Test(Description = "Insert and get")]
        public void TestInsertAndGet()
        {
            TEntity L_Entity = Create();
            Save(L_Entity);
            TEntity L_ReadEntity = Get(KeyGetter(L_Entity));
            Expect(AreSame(L_Entity, L_ReadEntity), Is.True);
        }

        [Test]
        public void TestUpdateAndGet()
        {
            TEntity L_Entity = Create();
            TEntity L_Entity_Modified = Create();
            L_Entity_Modified.PK = L_Entity.PK;
            Save(L_Entity_Modified);

            TEntity L_Read = Get(KeyGetter(L_Entity));
            Expect(AreSame(L_Entity_Modified, L_Read), Is.True);
        }

        [Test]
        public void TestInsertAndGetAll()
        {
            int L_Countb = GetAll().Count();
            TEntity L_Entity1 = Create();
            TEntity L_Entity2 = Create();
            TEntity L_Entity_Modified = Create();
            L_Entity_Modified.PK = L_Entity1.PK;
            L_Entity_Modified.AlreadyPersisted = true;
            Save(L_Entity1);
            Save(L_Entity2);
            Save(L_Entity_Modified);

            int L_Counta = GetAll().Count();
            Expect(L_Counta - L_Countb, Is.EqualTo(2));
        }

        [Test]
        public void TestInsertAndDelete()
        {
            int L_Countb = GetAll().Count();
            TEntity L_Entity1 = Create();
            TEntity L_Entity2 = Create();
            Save(L_Entity1);
            Save(L_Entity2);
            Delete(KeyGetter(L_Entity1));

            int L_Counta = GetAll().Count();
            Expect(L_Counta - L_Countb, Is.EqualTo(1));

            TEntity L_Read = Get(KeyGetter(L_Entity1));
            Expect(L_Read, Is.Null);
        }
    }
}